import 'dart:async';

import 'package:flutter/foundation.dart';
import 'package:flutter/rendering.dart';
import 'package:flutter/widgets.dart';
import 'package:flutter_hooks/flutter_hooks.dart';
import 'package:flutter_suspense/src/notifier.dart';
import 'package:functional_widget_annotation/functional_widget_annotation.dart';
import 'package:provider/provider.dart';
part 'suspense.dart';
part 'framework.g.dart';

class _Pair {
  final dynamic future;
  final BuildContext context;

  _Pair(this.future, this.context);

  operator ==(Object o) =>
      o is _Pair && o.future == future && o.context == context;

  @override
  int get hashCode => future.hashCode;
}

class SuspenseController {
  SuspenseController() {
    _loading = createNotifier(false);
  }

  Notifier<bool> createNotifier(bool initial) {
    return Notifier(initial);
  }

  var _loadingCount = 0;
  Notifier<bool> _loading;
  Notifiable<bool> get loading => _loading;

  final _futures = Set<_Pair>();

  VoidCallback dispatchFuture(dynamic future, BuildContext context) {
    final pair = _Pair(future, context);
    if (_futures.add(pair)) {
      _loadingCount++;
      if (_loadingCount == 1) {
        _loading.push(true);
      }
      return () => _unsubscribeFuture(pair);
    }
    return null;
  }

  void _unsubscribeFuture(_Pair pair) {
    if (!_futures.remove(pair)) {
      throw StateError('listener already disposed');
    }
    _loadingCount--;
    if (_loadingCount == 0) {
      _loading.push(false);
    }
  }

  void dispose() {
    _loading.dispose();
  }
}

// TODO: hot-reload seems to something throw when wrapping the widget
// TODO: unmounting Suspense seems to dispose the controller before RenderSuspense
@hwidget
Widget suspense({@required Widget child, @required Widget fallback}) {
  final controller = useMemoized(() => SuspenseController());

  // TODO: find out why this sizedbox is necessary to correctly render Scaffold
  return SizedBox.expand(
    child: _SuspenseVisibility(
      controller: controller,
      fallback: _SuspenseVisibilityWrapper(
        child: HookBuilder(builder: (context) {
          final animationEnabled = useState(false);
          useEffect(
            () => controller.loading.addListener(
                (val) => WidgetsBinding.instance.addPostFrameCallback((_) {
                      animationEnabled.value = val;
                    })),
            [controller],
          );

          return TickerMode(
            enabled: animationEnabled.value,
            child: fallback,
          );
        }),
      ),
      child: _SuspenseVisibilityWrapper(
        child: Provider<SuspenseController>(
          value: controller,
          child: child,
        ),
      ),
    ),
  );
}

void _useSuspense(dynamic future, AsyncSnapshot result) {
  assert(future is Future || future is Stream);
  final context = useContext();
  final controller = Provider.of<SuspenseController>(context);

  useEffect(
    () {
      if (result.connectionState != ConnectionState.done &&
          result.connectionState != ConnectionState.active) {
        return controller.dispatchFuture(future, context);
      }
    },
    [future, controller, result.connectionState],
  );

  Hook.use(const _OverrideErrorHook());
  if (result.hasError) {
    throw result.error;
  }
  if (result.connectionState != ConnectionState.done &&
      result.connectionState != ConnectionState.active) {
    // throw a Future so that the onError override is able to differentiate
    // voluntary aborts from real problems.
    throw future;
  }
}

T useStreamSuspense<T>(Stream<T> stream) {
  final result = useStream(stream, preserveState: false);
  _useSuspense(stream, result);
  return result.data;
}

T useFutureSuspense<T>(Future<T> future) {
  final result = useFuture(future, preserveState: false);
  _useSuspense(future, result);
  return result.data;
}

class _OverrideErrorHook extends Hook<void> {
  const _OverrideErrorHook();

  @override
  _OverrideErrorHookState createState() => _OverrideErrorHookState();
}

class _OverrideErrorHookState extends HookState<void, _OverrideErrorHook> {
  static Widget _ignore(FlutterErrorDetails details) => const SizedBox();

  FlutterExceptionHandler onError;
  ErrorWidgetBuilder errorBuilder;

  @override
  void build(BuildContext context) {
    onError = FlutterError.onError;
    errorBuilder = ErrorWidget.builder;

    FlutterError.onError = _onError;
    ErrorWidget.builder = _ignore;
  }

  void _onError(FlutterErrorDetails details) {
    if (details.exception is! Future && details.exception is! Stream) {
      onError(details);
      ErrorWidget.builder = errorBuilder;
    }
  }

  @override
  void didBuild() {
    FlutterError.onError = onError;
    ErrorWidget.builder = errorBuilder;
  }
}
