import 'package:flutter_suspense/src/framework.dart';
import 'package:flutter_suspense/src/notifier.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:mockito/mockito.dart';

import 'shared.dart';

class _SuspenseController extends SuspenseController {
  _SuspenseController(this.mock);

  Notifier<bool> mock;

  @override
  Notifier<bool> createNotifier(bool intialValue) => mock;
}

void main() {
  group('suspenseController', () {
    test('controller is not loading by default', () {
      final listener = MockListener<bool>();
      final controller = SuspenseController();

      controller.loading.addListener(listener);
      verify(listener(false));
    });
    test('disposing controller disposes internal notifier', () {
      final notifier = MockNotifier<bool>();
      final controller = _SuspenseController(notifier);

      verifyZeroInteractions(notifier);
      controller.dispose();
      verify(notifier.dispose()).called(1);
      verifyNoMoreInteractions(notifier);
    });
    test('futures and context are not used when dispatched', () {
      final controller = SuspenseController();
      final future = MockFuture();
      final context = MockBuildContext();

      controller.dispatchFuture(future, context);

      verifyZeroInteractions(future);
      verifyZeroInteractions(context);
    });
    test('dispatching a future call loading listeners only once', () {
      final notifier = MockNotifier<bool>();
      final controller = _SuspenseController(notifier);

      clearInteractions(notifier);
      controller.dispatchFuture(Future.value(42), MockBuildContext());
      controller.dispatchFuture(Future.value(42), MockBuildContext());

      verify(notifier.push(true)).called(1);
      verifyNoMoreInteractions(notifier);
    });
    test('dispatching a future returns a disposer only once per context+future',
        () {
      final notifier = MockNotifier<bool>();
      final controller = _SuspenseController(notifier);
      final future = Future.value(42);
      final context = MockBuildContext();

      clearInteractions(notifier);
      expect(controller.dispatchFuture(future, context), isNotNull);
      expect(controller.dispatchFuture(future, context), isNull);

      final context2 = MockBuildContext();
      expect(controller.dispatchFuture(future, context2), isNotNull);
      expect(controller.dispatchFuture(future, context2), isNull);

      final future2 = Future.value(42);
      expect(controller.dispatchFuture(future2, context2), isNotNull);
      expect(controller.dispatchFuture(future2, context2), isNull);
    });
    test('calling disposer more than once throws', () {
      final controller = SuspenseController();
      final disposer =
          controller.dispatchFuture(Future.value(42), MockBuildContext());

      disposer();
      expect(disposer, throwsStateError);
    });

    test('unsubscribing to all futures resets loading to false', () {
      final controller = SuspenseController();
      final listener = MockListener<bool>();
      controller.loading.addListener(listener);

      final dispose1 =
          controller.dispatchFuture(Future.value(42), MockBuildContext());
      final dispose2 =
          controller.dispatchFuture(Future.value(42), MockBuildContext());
      clearInteractions(listener);

      verifyNever(listener(false));
      dispose1();
      verifyNever(listener(false));
      dispose2();
      verify(listener(false)).called(1);
      verifyNoMoreInteractions(listener);
    });
  });
}
