import 'dart:async';

import 'package:flutter/foundation.dart';
import 'package:flutter/widgets.dart';
import 'package:flutter_hooks/flutter_hooks.dart' hide Dispose;
import 'package:flutter_suspense/src/framework.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:mockito/mockito.dart';
import 'package:provider/provider.dart';

import 'shared.dart';

class MockBuild extends Mock {
  Widget call();
}

void main() {
  group('useFutureSuspense', () {
    testWidgets(
        'changing of controller disposes the previous future sub, and sub again with the new controller',
        (tester) async {
      final completer = Completer();
      final controller = MockSuspenseController();
      final dispose = Dispose();
      when(controller.dispatchFuture(completer.future, any))
          .thenReturn(dispose);

      await tester.pumpWidget(Provider<SuspenseController>(
        value: controller,
        child: HookBuilder(builder: (context) {
          useFutureSuspense(completer.future);
        }),
      ));

      clearInteractions(controller);

      final controller2 = MockSuspenseController();
      await tester.pumpWidget(Provider<SuspenseController>(
        value: controller2,
        child: HookBuilder(builder: (context) {
          useFutureSuspense(completer.future);
        }),
      ));

      final context = tester.element(find.byType(HookBuilder));

      verifyInOrder([
        dispose(),
        controller2.dispatchFuture(completer.future, context),
      ]);
    });
    testWidgets('null works', (tester) async {
      final future = Future<int>.value(null);

      await tester.pumpWidget(Provider<SuspenseController>(
        value: SuspenseController(),
        child: HookBuilder(builder: (context) {
          final res = useFutureSuspense(future);
          return Text(res.toString(), textDirection: TextDirection.ltr);
        }),
      ));

      expect(find.text('null'), findsNothing);

      await future;
      await tester.pump();
      expect(find.text('null'), findsOneWidget);
    });
    testWidgets(
        'override onError and errorBuilder + retablish them after build',
        (tester) async {
      final originOnError = FlutterError.onError;
      final originErrorBuilder = ErrorWidget.builder;
      addTearDown(() {
        FlutterError.onError = originOnError;
        ErrorWidget.builder = originErrorBuilder;
      });
      ErrorWidgetBuilder errorBuilder;
      ErrorWidgetBuilder errorBuilder2;
      FlutterExceptionHandler onError;
      FlutterExceptionHandler onError2;

      await tester.pumpWidget(Provider<SuspenseController>(
        value: SuspenseController(),
        child: HookBuilder(builder: (context) {
          try {
            useFutureSuspense(Future.value());
          } finally {
            onError = FlutterError.onError;
            errorBuilder = ErrorWidget.builder;
            try {
              useFutureSuspense(Future.value());
            } finally {
              onError2 = FlutterError.onError;
              errorBuilder2 = ErrorWidget.builder;
            }
          }
        }),
      ));
      // retablish values
      expect(FlutterError.onError, equals(originOnError));
      expect(ErrorWidget.builder, equals(originErrorBuilder));

      // override
      expect(errorBuilder, isNot(equals(originErrorBuilder)));
      expect(errorBuilder2, equals(errorBuilder));

      expect(onError, isNot(equals(originOnError)));
      expect(onError2, isNot(equals(onError)));
    });
    testWidgets('unresolved future throws without calling onError/ErrorBuilder',
        (tester) async {
      final originOnError = FlutterError.onError;
      final originErrorBuilder = ErrorWidget.builder;

      var onError = MockOnError();
      FlutterError.onError = onError;
      var errorBuilder = MockErrorBuilder();
      ErrorWidget.builder = errorBuilder;

      await tester.pumpWidget(Provider<SuspenseController>(
        value: SuspenseController(),
        child: HookBuilder(builder: (context) {
          useFutureSuspense(Future.value());
        }),
      ));
      FlutterError.onError = originOnError;
      ErrorWidget.builder = originErrorBuilder;

      verifyNoMoreInteractions(onError);
      verifyNoMoreInteractions(errorBuilder);
    });
    testWidgets(
        'throwing non-future bubbles up to default onError and errorbuilder',
        (tester) async {
      var future = Future.value();
      var controller = SuspenseController();

      final originOnError = FlutterError.onError;
      final originErrorBuilder = ErrorWidget.builder;

      Widget build() {
        return Provider<SuspenseController>(
          value: controller,
          child: HookBuilder(builder: (context) {
            useFutureSuspense(future);
            throw 42;
          }),
        );
      }

      // first build will throw do to unresolved future
      await tester.pumpWidget(build());
      await future;

      var onError = MockOnError();
      FlutterError.onError = onError;

      var errorBuilder = MockErrorBuilder();
      ErrorWidget.builder = errorBuilder;

      await tester.pumpWidget(build());
      FlutterError.onError = originOnError;
      ErrorWidget.builder = originErrorBuilder;

      final capture = verify(onError(captureAny))..called(1);
      expect((capture.captured.first as FlutterErrorDetails).exception, 42);

      verifyZeroInteractions(errorBuilder);
    });

    testWidgets('aborts when future is pending by throwing the future',
        (tester) async {
      final build = MockBuild();
      Future catched;
      final completer = Completer<int>();

      await tester.pumpWidget(Provider<SuspenseController>(
        value: SuspenseController(),
        child: HookBuilder(builder: (context) {
          try {
            useFutureSuspense(completer.future);
          } catch (err) {
            catched = err;
            rethrow;
          }
          build();
        }),
      ));

      expect(catched, completer.future);
      verifyZeroInteractions(build);
    });

    testWidgets('dispatches the future to SuspenseController', (tester) async {
      final completer = Completer<int>();
      final controller = MockSuspenseController();

      await tester.pumpWidget(Provider<SuspenseController>(
        value: controller,
        child: HookBuilder(builder: (context) {
          useFutureSuspense(completer.future);
        }),
      ));
      final context = tester.element(find.byType(HookBuilder));

      verify(controller.dispatchFuture(completer.future, context));
    });
    testWidgets("rebuilds without changing future don't redispatch it",
        (tester) async {
      final completer = Completer<int>();
      final controller = MockSuspenseController();

      await tester.pumpWidget(Provider<SuspenseController>(
        value: controller,
        child: HookBuilder(builder: (context) {
          useFutureSuspense(completer.future);
        }),
      ));
      clearInteractions(controller);

      await tester.pumpWidget(Provider<SuspenseController>(
        value: controller,
        child: HookBuilder(builder: (context) {
          useFutureSuspense(completer.future);
        }),
      ));
      verifyNoMoreInteractions(controller);
    });
    testWidgets("returns the future result when it completes", (tester) async {
      final completer = Completer<String>();

      await tester.pumpWidget(Provider<SuspenseController>(
        value: SuspenseController(),
        child: HookBuilder(builder: (context) {
          return Text(
            useFutureSuspense(completer.future),
            textDirection: TextDirection.ltr,
          );
        }),
      ));

      expect(find.byType(Text), findsNothing);

      completer.complete('42');
      await completer.future;
      await tester.pump();

      expect(find.text('42'), findsOneWidget);
    });
    testWidgets("Don't dispatch again when build finishes", (tester) async {
      final completer = Completer<String>();
      var controller = MockSuspenseController();

      await tester.pumpWidget(Provider<SuspenseController>(
        value: controller,
        child: HookBuilder(builder: (context) {
          useFutureSuspense(completer.future);
          return Container();
        }),
      ));
      clearInteractions(controller);

      completer.complete('42');
      await completer.future;
      await tester.pump();

      verifyNoMoreInteractions(controller);
    });
    testWidgets(
        "unsubscribe to futures and sub to the new one when future change",
        (tester) async {
      final completer = Completer<String>();
      final completer2 = Completer<String>();
      var controller = MockSuspenseController();
      final dispose = Dispose();
      final dispose2 = Dispose();

      when(controller.dispatchFuture(completer.future, any))
          .thenReturn(dispose);
      when(controller.dispatchFuture(completer2.future, any))
          .thenReturn(dispose2);

      await tester.pumpWidget(Provider<SuspenseController>(
        value: controller,
        child: HookBuilder(builder: (context) {
          useFutureSuspense(completer.future);
        }),
      ));

      final context = tester.element(find.byType(HookBuilder));
      clearInteractions(controller);
      verifyZeroInteractions(dispose);
      verifyZeroInteractions(dispose2);

      await tester.pumpWidget(Provider<SuspenseController>(
        value: controller,
        child: HookBuilder(builder: (context) {
          // new completer
          useFutureSuspense(completer2.future);
        }),
      ));

      verifyInOrder([
        dispose(),
        controller.dispatchFuture(completer2.future, context),
      ]);
      verifyZeroInteractions(dispose2);
      verifyNoMoreInteractions(controller);
    });
    testWidgets("unsubscribe to future when widget is removed from the tree",
        (tester) async {
      final completer = Completer<String>();
      var controller = MockSuspenseController();
      final dispose = Dispose();
      when(controller.dispatchFuture(completer.future, any))
          .thenReturn(dispose);

      await tester.pumpWidget(Provider<SuspenseController>(
        value: controller,
        child: HookBuilder(builder: (context) {
          useFutureSuspense(completer.future);
        }),
      ));

      clearInteractions(controller);
      verifyZeroInteractions(dispose);

      await tester.pumpWidget(Provider<SuspenseController>(
        value: controller,
        child: Container(),
      ));

      dispose();
      verifyNoMoreInteractions(controller);
    });
    testWidgets("when future fails, throws the error", (tester) async {
      final completer = Completer<String>();

      await tester.pumpWidget(Provider<SuspenseController>(
        value: SuspenseController(),
        child: HookBuilder(builder: (context) {
          useFutureSuspense(completer.future);
        }),
      ));

      completer.completeError('Hello');
      await Future.value();

      await tester.pump();

      expect(tester.takeException(), 'Hello');
    });
    testWidgets('can be used multiple times with different futures',
        (tester) async {
      final completer = Completer<int>();
      final completer2 = Completer<String>();

      var controller = MockSuspenseController();
      var dispose = Dispose();
      var dispose2 = Dispose();
      when(controller.dispatchFuture(completer.future, any))
          .thenReturn(dispose);
      when(controller.dispatchFuture(completer2.future, any))
          .thenReturn(dispose2);

      await tester.pumpWidget(Provider<SuspenseController>(
        value: controller,
        child: HookBuilder(builder: (context) {
          final res1 = useFutureSuspense(completer.future);
          final res2 = useFutureSuspense(completer2.future);

          return Text('$res1 $res2', textDirection: TextDirection.ltr);
        }),
      ));
      final context = tester.element(find.byType(HookBuilder));

      clearInteractions(controller);
      verifyZeroInteractions(dispose);

      completer.complete(42);
      await completer.future;
      await tester.pump();

      verifyInOrder([
        dispose(),
        controller.dispatchFuture(completer2.future, context),
      ]);
      verifyZeroInteractions(dispose2);
      verifyNoMoreInteractions(controller);

      completer2.complete('24');
      await completer2.future;
      await tester.pump();

      verifyNoMoreInteractions(dispose);
      dispose2();
      verifyNoMoreInteractions(controller);
      expect(find.text('42 24'), findsOneWidget);
    });
  });
}
